from dataclasses import dataclass
from typing import Optional

import torch
from datasets import Dataset
from torch.utils.data import DataLoader, SequentialSampler
from tqdm.auto import tqdm
from transformers import (
    AutoConfig,
    AutoTokenizer,
    LongformerForQuestionAnswering,
    set_seed,
)

from entity_linkings.trainer import EntityLinkingTrainer, TrainingArguments
from entity_linkings.utils import BaseSystemOutput, calculate_top1_accuracy

from ..base import RerankerBase, RetrieverBase
from .collator import CollatorForExtend
from .preprocessor import ExtendPreprocessor
from .utils import compute_metrics, select_indices


class EXTEND(RerankerBase):

    @dataclass
    class Config(RerankerBase.Config):
        '''EXTEND configuration
        Args:
            - model_name_or_path (str): Pre-trained model name or path for mention and entity encoders
            - add_nil_token (bool): Whether to add a special token for NIL entity
            - nil_token (str): Special token for NIL entity
            - max_context_length (int): Maximum sequence length for mention context
            - pooling (str): Pooling method for obtaining fixed-size representations ('cls' or 'mean')
            - batch_size (int): Batch size for encoding mentions and entities
        '''
        model_name_or_path: str = "allenai/longformer-large-4096"
        ent_start_token: str = "[START_ENT]"
        ent_end_token: str = "[END_ENT]"
        nil_token: str = "[NIL]"
        max_context_length: int = 128
        context_window_chars: int = 500
        candidate_separator: str = '*'
        attention_window: int = 64
        modify_global_attention: int = 2
        mode: str = "max-prod" # "max-prod" or "max-end", "max-start", "max"

    def __init__(self, retriever: RetrieverBase, config: Optional[Config] = None) -> None:
        super().__init__(retriever, config)
        self.tokenizer = AutoTokenizer.from_pretrained(self.config.model_name_or_path)
        model_config = AutoConfig.from_pretrained(
            self.config.model_name_or_path,
            attention_window=self.config.attention_window
        )
        self.model = LongformerForQuestionAnswering.from_pretrained(self.config.model_name_or_path, config=model_config)
        special_tokens = [self.config.ent_start_token, self.config.ent_end_token, self.config.nil_token]
        self.tokenizer.add_special_tokens({"additional_special_tokens": special_tokens})
        if self.model.config.vocab_size != len(self.tokenizer):
            self.model.resize_token_embeddings(len(self.tokenizer))
        self.preprocessor = ExtendPreprocessor(
            tokenizer = self.tokenizer,
            dictionary = self.retriever.dictionary,
            ent_start_token=self.config.ent_start_token,
            ent_end_token=self.config.ent_end_token,
            candidate_separator=self.config.candidate_separator,
            max_context_length=self.config.max_context_length,
        )

    def train(
            self,
            train_dataset: Dataset,
            eval_dataset: Optional[Dataset] = None,
            num_candidates: int = 30,
            training_args: Optional[TrainingArguments] = None
        ) -> dict[str, float]:
        if training_args is None:
            training_args = TrainingArguments()
        set_seed(training_args.seed)

        train_candidates = self.retriever.retrieve_candidates(
            train_dataset,
            only_negative=True,
            top_k=num_candidates,
            batch_size=training_args.per_device_eval_batch_size,
        )

        train_dataset = self.preprocessor.dataset_preprocess(train_dataset, train_candidates, train=True)
        if eval_dataset is not None:
            eval_candidates = self.retriever.retrieve_candidates(
                eval_dataset,
                only_negative=True,
                top_k=num_candidates,
                batch_size=training_args.per_device_eval_batch_size,
            )
            eval_dataset = self.preprocessor.dataset_preprocess(eval_dataset, eval_candidates, train=True)

        trainer = EntityLinkingTrainer(
            model=self.model,
            args=training_args,
            data_collator=CollatorForExtend(
                self.tokenizer,
                modify_global_attention=self.config.modify_global_attention,
            ),
            train_dataset=train_dataset,
            eval_dataset=eval_dataset,
            compute_metrics=compute_metrics,
        )
        checkpoint = None
        if training_args.resume_from_checkpoint is not None:
            checkpoint = training_args.resume_from_checkpoint
        results = trainer.train(resume_from_checkpoint=checkpoint)
        trainer.log_metrics("train", results.metrics)
        if training_args.output_dir is not None:
            self.model.save_pretrained(training_args.output_dir)
            self.tokenizer.save_pretrained(training_args.output_dir)
            trainer.save_state()
            trainer.save_metrics("train", results.metrics)
        return results

    @torch.no_grad()
    def evaluate(self, dataset: Dataset, num_candidates: int = 30, batch_size: int = 32, **args: int) -> dict[str, float]:
        candidates = self.retriever.retrieve_candidates(
            dataset, only_negative=False, top_k=num_candidates, batch_size=batch_size,
        )

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)

        processed_dataset = self.preprocessor.dataset_preprocess(dataset, candidates, train=False)
        dataloader = DataLoader(
            processed_dataset,
            batch_size=batch_size,
            sampler=SequentialSampler(processed_dataset),
            collate_fn=CollatorForExtend(
                self.tokenizer,
                modify_global_attention=self.config.modify_global_attention,
            ),
        )

        pbar  = tqdm(total=len(dataloader), desc='Evaluate')
        start_probs, end_probs = [], []
        for batch in dataloader:
            pbar.update()
            batch = batch.to(device)
            outputs = self.model(**batch)
            start_logits = outputs.start_logits
            end_logits = outputs.end_logits
            start_probs.extend(torch.softmax(start_logits, dim=-1).tolist()) # (batch_size, seq_length)
            end_probs.extend(torch.softmax(end_logits, dim=-1).tolist()) # ( batch_size, seq_length)
        pbar.close()

        assert len(start_probs) == len(processed_dataset)
        assert len(end_probs) == len(processed_dataset)
        num_golds = len(processed_dataset)
        num_corrects = 0
        for data, start_prob, end_prob in zip(processed_dataset, start_probs, end_probs):
            prob = torch.tensor([start_prob, end_prob])  # (2, seq_length)
            start_index, end_index = select_indices(
                mode=self.config.mode,
                possible_indices=data["candidates_offsets"],
                classification_probabilities=prob,
            )
            pred_index = data['candidates_offsets'].index([start_index, end_index])
            pred = data["candidates"][pred_index]
            if pred in data["labels"]:
                num_corrects += 1
        metric = calculate_top1_accuracy(num_corrects, num_golds)
        return metric

    @torch.no_grad()
    def predict(self, sentence: str, spans: list[tuple[int, int]], num_candidates: int = 30) -> list[BaseSystemOutput]:
        predictions = self.retriever.predict(sentence, spans=spans, top_k=num_candidates)
        candidates = [[cand_id.id for cand_id in c] for c in predictions]

        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model.eval()
        self.model.to(device)

        all_result = []
        collator = CollatorForExtend(self.tokenizer, modify_global_attention=self.config.modify_global_attention)
        for i, (b, e) in enumerate(spans):
            processed_text = self.preprocessor.process_context(sentence, b, e, candidates[i])
            encodings = collator([processed_text])
            encodings = encodings.to(device)
            outputs = self.model(**encodings)
            start_logits = outputs.start_logits
            end_logits = outputs.end_logits
            scores = torch.stack([start_logits, end_logits], dim=0) # (2, batch_size, seq_length)
            probabilities = torch.softmax(scores, dim=-1) # (2, batch_size, seq_length)
            start_index, end_index = select_indices(
                mode=self.config.mode,
                possible_indices=processed_text['candidates_offsets'],
                classification_probabilities=probabilities[:, 0],
            )
            pred = processed_text['candidates_offsets'].index((start_index, end_index))
            top1_entity = self.dictionary(candidates[i][pred])
            all_result.append(BaseSystemOutput(query=sentence[b:e], start=b, end=e, id=top1_entity['id']))
        return all_result



import json
import os
from argparse import ArgumentParser

from datasets import DownloadManager
from rdflib import Graph, Namespace, URIRef
from tqdm.auto import tqdm

from appendix.data_convert.utils import WikiMapper, read_text_file

_URLs = {
    "train": "https://raw.githubusercontent.com/anuzzolese/oke-challenge-2016/refs/heads/master/GoldStandard_sampleData/task1/dataset_task_1.ttl",
    "test": "https://raw.githubusercontent.com/anuzzolese/oke-challenge-2016/refs/heads/master/evaluation-data/task1/evaluation-dataset-task1.ttl"
}


def get_documents(g: Graph, prefix: Namespace) -> dict[str, str]:
    documents = {}
    for s, p, o in g:
        if str(p) == str(prefix.isString):
            doc_id = str(s).split('#')[0].split('/')[-1]
            documents[str(doc_id)] = str(o)
    return documents


def main(finder: WikiMapper, output_dir: str) -> None:
    output_dir_data = os.path.join(output_dir, 'oke2016')
    os.makedirs(output_dir_data, exist_ok=True)
    print(f"Output directory data: {output_dir_data}")

    for split, url in _URLs.items():
        data_path = DownloadManager().download_and_extract(url)
        raw_text = read_text_file(data_path)

        prefix = Namespace('http://persistence.uni-leipzig.org/nlp2rdf/ontologies/nif-core#')
        g = Graph()
        g.parse(data=raw_text, format='turtle')

        raw_documents = get_documents(g, prefix)
        with open(os.path.join(output_dir_data, f"{split}.jsonl"), 'w', encoding='utf-8') as f:
            pbar = tqdm(total=len(raw_documents), desc="Processing documents")
            for doc_id, raw_doc in raw_documents.items():
                pbar.update()
                entities = []
                anchors = [s for s in g.subjects(predicate=prefix.anchorOf) if f'http://www.ontologydesignpatterns.org/data/oke-challenge/task-1/{doc_id}#' in str(s)]
                for anchor in anchors:
                    start = int(g.value(subject=anchor, predicate=prefix.beginIndex))
                    end = int(g.value(subject=anchor, predicate=prefix.endIndex))
                    reference = str(g.value(subject=anchor, predicate=URIRef("http://www.w3.org/2005/11/its/rdf#taIdentRef")))
                    wikiname = reference.split("/")[-1]
                    wikiname = wikiname[len("sentence-"):] if wikiname.startswith("sentence-") else wikiname
                    res = finder.find_by_title(wikiname)
                    page_id = res["page_id"] if res is not None and res["page_id"] is not None else "-1"
                    entities.append({"start": start, "end": end, "label": [str(page_id)]})
                example = {
                    "subset": "oke2016",
                    "id": doc_id,
                    "text": raw_doc,
                    "entities": entities
                }
                f.write(f"{json.dumps(example, ensure_ascii=False)}\n")
            pbar.close()


if __name__ == "__main__":
    parser = ArgumentParser(description="OKE2016 Entity Disambiguation Script")
    parser.add_argument("--output_dir", "-o", type=str, default='processed_data', help="Path to the output text file")
    parser.add_argument("--wiki_map_file", '-w', type=str, default='wikipedia_to_wikidata.duckdb', help="Path to the wiki mapping file.")
    args = parser.parse_args()

    finder = WikiMapper(args.wiki_map_file)
    main(finder, output_dir=args.output_dir)
    finder.close()



# import json
import json
import os
from logging import getLogger
from typing import Optional

import faiss
import numpy as np
import torch
from torch.utils.data import DataLoader, SequentialSampler
from tqdm.auto import tqdm
from transformers import PreTrainedTokenizer

from entity_linkings.data_utils import CollatorBase, EntityDictionary

from ..base import IndexerBase
from .encoder import DualBERTModel

logger = getLogger(__name__)


class DenseRetriever(IndexerBase):
    def __init__(
        self,
        model: DualBERTModel,
        tokenizer: PreTrainedTokenizer,
        dictionary: EntityDictionary,
        batch_size: int = 16,
        metric: str = "cosine",
        use_hnsw: bool = False,
        n_hubs: int = 10,
        fp16: bool = False
    ) -> None:
        super().__init__(dictionary)
        self.model = model
        self.tokenizer = tokenizer
        self.vector_size = self.model.hidden_size
        self.metric = metric
        self.use_hnsw = use_hnsw
        self.n_hubs = n_hubs
        self.fp16 = fp16
        self.device = torch.device("cuda") if torch.cuda.is_available() else torch.device("cpu")
        self.batch_size = batch_size

    def _initialize(self) -> None:
        self.meta_ids_to_keys: dict[int, str] = {}
        if self.metric not in ["cosine", "inner_product", "euclidean"]:
            raise NotImplementedError(f"{self.metric} is not supported")
        if self.use_hnsw:
            if self.metric == 'cosine' or self.metric == 'inner_product':
                self.index = faiss.IndexHNSWFlat(self.vector_size, self.n_hubs, faiss.METRIC_INNER_PRODUCT)
            else:
                self.index = faiss.IndexHNSWFlat(self.vector_size, self.n_hubs, faiss.METRIC_L2)
        else:
            if self.metric == 'cosine' or self.metric == 'inner_product':
                self.index = faiss.IndexFlatIP(self.vector_size)
            else:
                self.index = faiss.IndexFlatL2(self.vector_size)

    @torch.no_grad()
    def build_index(self, index_path: str) -> None:
        if os.path.exists(os.path.join(index_path, "index.dpr")) and os.path.exists(os.path.join(index_path, "meta.json")):
            logger.info(f"Loading index from {index_path}")
            self.load(index_path)
        else:
            self._initialize()
            self.meta_ids_to_keys = {k: idx for idx, k in self.dictionary.id_to_index.items()}
            self.model.eval()
            self.model.to(self.device)

            dataloader = DataLoader(
                self.dictionary,
                collate_fn=CollatorBase(self.tokenizer),
                batch_size=self.batch_size,
                sampler=SequentialSampler(self.dictionary)
            )
            pbar = tqdm(total=len(dataloader), desc='Build Index')
            for batch in dataloader:
                pbar.update()
                batch = batch.to(self.device)
                if self.fp16:
                    with torch.autocast(device_type=self.device.type):
                        entity_embedding = self.model.encode_candidate(**batch).to('cpu').detach().numpy().copy()
                else:
                    entity_embedding = self.model.encode_candidate(**batch).to('cpu').detach().numpy().copy()
                if self.metric == 'cosine':
                    faiss.normalize_L2(entity_embedding)
                self.index.add(entity_embedding)
            pbar.close()

    @torch.no_grad()
    def search_knn(self, query: str|list[str], top_k: int, ignore_ids: Optional[list[str]|list[list[str]]] = None) -> tuple[np.ndarray, list[list[str]]]:
        self.model.eval()
        self.model.to(self.device)
        model_inputs = self.tokenizer(query, padding=True, truncation=True, return_tensors="pt")
        model_inputs = model_inputs.to(self.device)
        if self.fp16:
            with torch.autocast(device_type=self.device.type):
                query_embed = self.model.encode_mention(**model_inputs).to('cpu').numpy()
        else:
            query_embed = self.model.encode_mention(**model_inputs).to('cpu').numpy()
        if top_k <= 0:
            raise RuntimeError("K is zero or under zero.")
        if top_k > len(self.meta_ids_to_keys):
            top_k = len(self.meta_ids_to_keys)
            logger.warning(f"K is over size of dictionary. K is modified to size of dictionary to {len(self.meta_ids_to_keys)}")

        if self.metric == 'cosine':
            faiss.normalize_L2(query_embed)

        additional_top_k = 0
        if ignore_ids is not None:
            if isinstance(ignore_ids[0], list):
                additional_top_k = max([len(ids) for ids in ignore_ids])
            else:
                additional_top_k = len(ignore_ids)
        scores, results = self.index.search(query_embed, k=top_k+additional_top_k)

        indices_keys = []
        for i in range(len(results)):
            if ignore_ids is None:
                indices_keys.append([self.meta_ids_to_keys[ind] for ind in results[i]])
                continue
            candidate_ids = []
            for j in results[i]:
                key = self.meta_ids_to_keys[j]
                if key not in ignore_ids[i]:
                    candidate_ids.append(key)
            indices_keys.append(candidate_ids[:top_k])
        return scores, indices_keys

    def save_index(self, index_path: str, ensure_ascii: bool = False) -> None:
        logger.info("Serializing index to %s", index_path)
        if not os.path.isdir(index_path):
            os.makedirs(index_path, exist_ok=True)
        index_file = os.path.join(index_path, "index.dpr")
        meta_file = os.path.join(index_path, "meta.json")
        faiss.write_index(self.index, index_file)
        json.dump(self.meta_ids_to_keys, open(meta_file, 'w'), ensure_ascii=ensure_ascii)

    def load(self, index_path: str) -> None:
        self._initialize()
        if not os.path.exists(index_path):
            raise FileNotFoundError(f"{index_path} is not found.")
        index_file = os.path.join(index_path, "index.dpr")
        meta_file = os.path.join(index_path, "meta.json")
        logger.info("Loading index from %s", index_file)
        self.index = faiss.read_index(index_file)
        self.meta_ids_to_keys.update({int(k): v for k, v in json.load(open(meta_file)).items()})
        logger.info(
            "Loaded index of type %s and size %d", type(self.index), len(self.meta_ids_to_keys)
        )

    def __len__(self) -> int:
        return len(self.meta_ids_to_keys)



from importlib.resources import files

from datasets import load_dataset

import assets as test_data
from entity_linkings import load_dictionary

from .utils import (
    cut_context_window,
    filter_nil_entities,
    sample_range_excluding,
    truncate_around_mention,
)

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']

def test_filter_nil_entities() -> None:
    filtered_dataset = filter_nil_entities(dataset, dictionary)
    for example in filtered_dataset:
        for entity in example["entities"]:
            for label in entity["label"]:
                assert dictionary(label) != dictionary.nil_id
    assert len(filtered_dataset) == len(dataset)
    assert len(filtered_dataset) == 8


def test_sample_range_excluding() -> None:
    n, k = 10, 3
    excluding = [2, 5, 7]
    sampled = sample_range_excluding(n, k, excluding)
    assert len(sampled) == k
    for s in sampled:
        assert s not in excluding
        assert 0 <= s < n

def test_cut_context_window() -> None:
    text = "This is a sample text for testing the cut_context_window function."
    start, end = 10, 16  # "sample"
    context_window_chars = 30

    context, new_start, new_end = cut_context_window(text, start, end, context_window_chars)

    assert len(context) <= context_window_chars
    assert context[new_start:new_end] == "sample"

    # Test edge case where mention is at the start
    text_start = "sample text for testing the cut_context_window function."
    start_start, end_start = 0, 6  # "sample"
    context_start, new_start_start, new_end_start = cut_context_window(text_start, start_start, end_start, context_window_chars)
    assert context_start[new_start_start:new_end_start] == "sample"

    # Test edge case where mention is at the end
    text_end = "This is a sample text for testing the cut_context_window function sample"
    start_end, end_end = len(text_end) - 6, len(text_end)  # "sample"
    context_end, new_start_end, new_end_end = cut_context_window(text_end, start_end, end_end, context_window_chars)
    assert context_end[new_start_end:new_end_end] == "sample"


def test_truncate_around_mention() -> None:
    tokens_ids = list(range(100))
    max_token_length = 50
    ent_start_idx = 45
    ent_end_idx = 55

    truncated_ids, new_start, new_end = truncate_around_mention(
        tokens_ids, max_token_length, ent_start_idx, ent_end_idx
    )

    assert len(truncated_ids) <= max_token_length
    assert new_start >= 0 and new_end <= len(truncated_ids)
    assert new_end - new_start == ent_end_idx - ent_start_idx
    assert truncated_ids[new_start:new_end] == tokens_ids[ent_start_idx:ent_end_idx]



from .collator import CollatorBase
from .entity_dictionary import EntityDictionary
from .preprocessor import Preprocessor, preprocess
from .utils import (
    cut_context_window,
    filter_nil_entities,
    sample_range_excluding,
    truncate_around_mention,
)

__all__ = [
    "CollatorBase",
    "EntityDictionary",
    "filter_nil_entities",
    "cut_context_window",
    "sample_range_excluding",
    "truncate_around_mention",
    "preprocess",
    "Preprocessor",
]



from importlib.resources import files

import pytest
from datasets import load_dataset
from transformers import AutoTokenizer

import assets as test_data
from entity_linkings import get_retrievers, load_dictionary

from .preprocessor import Preprocessor

MODELS = [
    "google-bert/bert-base-uncased",
    "FacebookAI/xlm-roberta-base",
    "microsoft/deberta-v3-base",
    "FacebookAI/roberta-base",
    "answerdotai/ModernBERT-base",
    "google-t5/t5-small",
]

dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dataset = load_dataset("json", data_files={"test": dataset_path})['test']

retriever_cls = get_retrievers("bm25")
retriever = retriever_cls(dictionary=dictionary)
candidates_ids = retriever.retrieve_candidates(dataset, top_k=3, only_negative=True)

class TestPreprocessor:
    @pytest.mark.parametrize("model_name", MODELS)
    def test_init(self, model_name: str) -> None:
        tokenizer = AutoTokenizer.from_pretrained(model_name)
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENT_START]", "[ENT_END]", "[ENTITY]"]})
        preprocessor = Preprocessor(
            tokenizer=tokenizer,
            ent_start_token="[ENT_START]",
            ent_end_token="[ENT_END]",
            entity_token="[ENTITY]",
            max_context_length=512,
            max_candidate_length=32,
            context_window_chars=100,
        )
        assert preprocessor.ent_start_token == "[ENT_START]"
        assert preprocessor.ent_end_token == "[ENT_END]"
        assert preprocessor.entity_token == "[ENTITY]"
        assert len(preprocessor.prefix_ids) == 0 if model_name.startswith("google-t5") else 1
        assert len(preprocessor.suffix_ids) == 1
        assert preprocessor.max_context_length == 512 - len(preprocessor.prefix_ids) - len(preprocessor.suffix_ids)
        assert preprocessor.offset_correction == len(preprocessor.prefix_ids)
        assert preprocessor.max_candidate_length == 32
        assert preprocessor.context_window_chars == 100

    def test_dictionary_preprocess(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENT_START]", "[ENT_END]", "[ENTITY]"]})
        preprocessor = Preprocessor(tokenizer=tokenizer, ent_start_token="[ENT_START]", ent_end_token="[ENT_END]", entity_token="[ENTITY]", max_candidate_length=32)
        processed_dictionary = preprocessor.dictionary_preprocess(dictionary)
        for entry in processed_dictionary:
            assert "encoding" in entry
            assert isinstance(entry["encoding"], list)
            assert len(entry["encoding"]) <= 32

    def test_process_context(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENT_START]", "[ENT_END]", "[ENTITY]"]})
        preprocessor = Preprocessor(tokenizer=tokenizer, ent_start_token="[ENT_START]", ent_end_token="[ENT_END]", entity_token="[ENTITY]", max_context_length=128, context_window_chars=200)
        text = dataset['text'][0]
        entity = dataset['entities'][0][0]
        encodings = preprocessor.process_context(text, entity['start'], entity['end'])
        assert "input_ids" in encodings
        assert "attention_mask" in encodings
        start_pos = encodings['input_ids'].index(preprocessor.start_marker_id)
        end_pos = encodings['input_ids'].index(preprocessor.end_marker_id)
        mention_ids = encodings['input_ids'][start_pos: end_pos + 1]
        mention_text = text[entity['start']: entity['end']].lower()
        decoded_mention = tokenizer.decode(mention_ids, skip_special_tokens=False).replace(" ", "")
        expected_mention = f"{preprocessor.ent_start_token}{mention_text}{preprocessor.ent_end_token}".replace(" ", "")
        assert decoded_mention == expected_mention
        assert len(encodings["input_ids"]) == len(encodings["attention_mask"])
        assert len(encodings["input_ids"]) <= preprocessor.max_context_length

    def test_data_flatten(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENT_START]", "[ENT_END]", "[ENTITY]"]})
        preprocessor = Preprocessor(tokenizer=tokenizer, ent_start_token="[ENT_START]", ent_end_token="[ENT_END]", entity_token="[ENTITY]", max_candidate_length=32)
        flattened_dataset = preprocessor.data_flatten(dataset)
        assert len(flattened_dataset) == 8
        assert len(dataset) == 8
        assert len(set([d['id'] for d in flattened_dataset])) == 7

    def test_dataset_preprocess(self) -> None:
        tokenizer = AutoTokenizer.from_pretrained(MODELS[0])
        tokenizer.add_special_tokens({"additional_special_tokens": ["[ENT_START]", "[ENT_END]", "[ENTITY]"]})
        preprocessor = Preprocessor(
            tokenizer=tokenizer,
            ent_start_token="[ENT_START]",
            ent_end_token="[ENT_END]",
            entity_token="[ENTITY]",
            max_context_length=128,
            max_candidate_length=32,
            context_window_chars=200,
        )
        processed_dataset = preprocessor.dataset_preprocess(dataset, candidates_ids)
        assert len(processed_dataset) == 8
        for processed in processed_dataset:
            assert "input_ids" in processed
            assert "attention_mask" in processed
            assert "start_positions" in processed
            assert "end_positions" in processed
            assert "candidates" in processed
            assert "labels" in processed

        processed_dataset = preprocessor.dataset_preprocess(dataset)
        assert len(processed_dataset) == 8
        for processed in processed_dataset:
            assert "input_ids" in processed
            assert "attention_mask" in processed
            assert "start_positions" in processed
            assert "end_positions" in processed
            assert "candidates" not in processed
            assert "labels" in processed



import tempfile
from importlib.resources import files

import numpy as np
import pytest
from bm25s.hf import BM25HF
from datasets import load_dataset

import assets as test_data
from entity_linkings import load_dictionary

from .indexer import BM25Indexer

dataset_path = str(files(test_data).joinpath("dataset_toy.jsonl"))
dictionary_path = str(files(test_data).joinpath("dictionary_toy.jsonl"))
dictionary = load_dictionary(dictionary_path)
dataset = load_dataset("json", data_files={"test": dataset_path})['test']

MODEL = ["google-bert/bert-base-uncased", None]

class TestBM25Indexer:
    @pytest.mark.parametrize("model_name", MODEL)
    def test___init__(self, model_name: str | None) -> None:
        indexer = BM25Indexer(dictionary, subword_tokenizer=model_name)
        assert isinstance(indexer, BM25Indexer)
        assert indexer.dictionary == dictionary
        if model_name is None:
            assert indexer.tokenize_func == indexer.whitespace_tokenize
        else:
            assert indexer.tokenize_func == indexer.subword_tokenize

    @pytest.mark.parametrize("model_name", MODEL)
    def test_initialize(self, model_name: str | None) -> None:
        indexer = BM25Indexer(dictionary, subword_tokenizer=model_name)
        indexer._initialize()
        assert isinstance(indexer.index, BM25HF)
        assert isinstance(indexer.meta_ids_to_keys, dict)
        assert len(indexer.meta_ids_to_keys) == 0

    @pytest.mark.parametrize("model_name", MODEL)
    def test_build_index(self, model_name: str | None) -> None:
        indexer = BM25Indexer(dictionary, subword_tokenizer=model_name)
        indexer.build_index()
        assert len(indexer) == len(dictionary)
        assert len(list(indexer.meta_ids_to_keys.keys())) == len(dictionary)

    @pytest.mark.parametrize("top_k", [0, 2, 5, 10])
    def test_search_knn(self, top_k: int) -> None:
        indexer = BM25Indexer(dictionary)
        indexer.build_index()
        queries = ["Microsoft", "Apple"]
        if top_k <= 0:
            with pytest.raises(RuntimeError) as re:
                indexer.search_knn(queries, top_k)
            assert isinstance(re.value, RuntimeError)
            assert str(re.value) == "K is zero or under zero."
        else:
            distances, indices = indexer.search_knn(queries, top_k)
            assert isinstance(distances, np.ndarray) and isinstance(indices, list)
            if top_k > len(dictionary):
                assert distances.shape[0] == len(indices) == 2
                assert distances.shape[1] == len(indices[0]) == len(dictionary)
            else:
                assert distances.shape[0] == len(indices) == 2
                assert distances.shape[1] == len(indices[0]) == top_k

    @pytest.mark.parametrize("top_k", [2, 4])
    def test_search_knn_negatives(self, top_k: int) -> None:
        indexer = BM25Indexer(dictionary)
        indexer.build_index()
        for example in dataset:
            if not example["entities"]:
                continue
            ignore_ids = [entity["label"] for entity in example["entities"]]
            queries = [example["text"][ent["start"]: ent["end"]] for ent in example["entities"]]
            _, indices = indexer.search_knn(queries, top_k, ignore_ids=ignore_ids)
            for i, inds in enumerate(indices):
                assert len(inds) == top_k
                for ind in inds:
                    assert ind not in ignore_ids[i]

    def test_save_and_load(self) -> None:
        indexer = BM25Indexer(dictionary)
        indexer.build_index()
        with tempfile.TemporaryDirectory() as tmpdir:
            indexer.save_index(tmpdir)
            loaded_indexer = BM25Indexer(dictionary)
            loaded_indexer.build_index(index_path=tmpdir)
            assert indexer.dictionary and loaded_indexer.dictionary
            assert indexer.meta_ids_to_keys == loaded_indexer.meta_ids_to_keys

    def test_len(self) -> None:
        indexer = BM25Indexer(dictionary)
        indexer.build_index()
        assert len(indexer) == len(dictionary)
